import numpy as np
from .. import pixel_formats
from ..core import Format


class V210(Format):
    """V210 interleaved 10bit yuv 422 format."""
    @staticmethod
    def identifier():
        return "v210"

    @staticmethod
    def chroma_subsampling():
        return 2, 1

    @staticmethod
    def bitdepth():
        return 10

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('word0', "<u4"),
              ('word1', "<u4"),
              ('word2', "<u4"),
              ('word3', "<u4")],
             ((self._height * self._width) // 6))
        ])

    def unpack(self, data):
        word0 = data['frame']['word0']
        cr0 = np.right_shift(np.bitwise_and(word0, (0x3ff << 20)), 20).astype(np.uint16)
        y0 = np.right_shift(np.bitwise_and(word0, (0x3ff << 10)), 10).astype(np.uint16)
        cb0 = np.bitwise_and(word0, 0x3ff).astype(np.uint16)

        word1 = data['frame']['word1']
        y2 = np.right_shift(np.bitwise_and(word1, (0x3ff << 20)), 20).astype(np.uint16)
        cb1 = np.right_shift(np.bitwise_and(word1, (0x3ff << 10)), 10).astype(np.uint16)
        y1 = np.bitwise_and(word1, 0x3ff).astype(np.uint16)

        word2 = data['frame']['word2']
        cb2 = np.right_shift(np.bitwise_and(word2, (0x3ff << 20)), 20).astype(np.uint16)
        y3 = np.right_shift(np.bitwise_and(word2, (0x3ff << 10)), 10).astype(np.uint16)
        cr1 = np.bitwise_and(word2, 0x3ff).astype(np.uint16)

        word3 = data['frame']['word3']
        y5 = np.right_shift(np.bitwise_and(word3, (0x3ff << 20)), 20).astype(np.uint16)
        cr2 = np.right_shift(np.bitwise_and(word3, (0x3ff << 10)), 10).astype(np.uint16)
        y4 = np.bitwise_and(word3, 0x3ff).astype(np.uint16)

        y = np.stack((y0, y1, y2, y3, y4, y5), 2).reshape((-1, self._height, self._width))
        u = np.stack((cb0, cb1, cb2), 2).reshape((-1, self._height, self._width // 2))
        v = np.stack((cr0, cr1, cr2), 2).reshape((-1, self._height, self._width // 2))

        return y, u, v

    def pack(self, yuv):
        y, u, v = yuv
        y = y.reshape((-1, (self._height * self._width) // 6, 6))
        u = u.reshape((-1, (self._height * self._width) // 6, 3))
        v = v.reshape((-1, (self._height * self._width) // 6, 3))
        y0, y1, y2, y3, y4, y5 = y[:, :, 0], y[:, :, 1], y[:, :, 2], y[:, :, 3], y[:, :, 4], y[:, :, 5]
        cb0, cb1, cb2 = u[:, :, 0], u[:, :, 1], u[:, :, 2]
        cr0, cr1, cr2 = v[:, :, 0], v[:, :, 1], v[:, :, 2]

        data = np.empty(y.shape[0], dtype=self.dtype)
        data['frame']['word0'] = np.left_shift(cr0.astype(np.uint32), 20) \
                                 + np.left_shift(y0.astype(np.uint32), 10) \
                                 + cb0.astype(np.uint32)
        data['frame']['word1'] = np.left_shift(y2.astype(np.uint32), 20) \
                                 + np.left_shift(cb1.astype(np.uint32), 10) \
                                 + y1.astype(np.uint32)
        data['frame']['word2'] = np.left_shift(cb2.astype(np.uint32), 20) \
                                 + np.left_shift(y3.astype(np.uint32), 10) \
                                 + cr1.astype(np.uint32)
        data['frame']['word3'] = np.left_shift(y5.astype(np.uint32), 20) \
                                 + np.left_shift(cr2.astype(np.uint32), 10) \
                                 + y4.astype(np.uint32)
        return data


pixel_formats.register(V210)



from abc import ABC
import numpy as np
from .. import pixel_formats
from ..core import Format


class _Interleaved422Base(Format, ABC):
    """Base for all interleaved 422 formats."""
    @staticmethod
    def chroma_subsampling():
        return 2, 1

    def unpack(self, data):
        y = np.stack((data['frame']['y0'],
                      data['frame']['y1']), 3).reshape((-1, self._height, self._width))
        u = data['frame']['u0']
        v = data['frame']['v0']
        return y, u, v

    def pack(self, yuv):
        y, u, v = yuv
        data = np.empty(y.shape[0], dtype=self.dtype)
        y = y.reshape((-1, self._height, self._width // 2, 2))
        y0, y1 = y[:, :, :, 0], y[:, :, :, 1]
        data['frame']['y0'][:] = y0
        data['frame']['u0'][:] = u
        data['frame']['y1'][:] = y1
        data['frame']['v0'][:] = v
        return data


class YUYV422(_Interleaved422Base):
    @staticmethod
    def identifier():
        return "yuyv422"

    @staticmethod
    def bitdepth():
        return 8

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y0', "<u1"),
              ('u0', "<u1"),
              ('y1', "<u1"),
              ('v0', "<u1")],
             (self._height, self._width // 2))
        ])


class UYVY422(_Interleaved422Base):
    @staticmethod
    def identifier():
        return "uyvy422"

    @staticmethod
    def bitdepth():
        return 8

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('u0', "<u1"),
              ('y0', "<u1"),
              ('v0', "<u1"),
              ('y1', "<u1")],
             (self._height, self._width // 2))
        ])


class YVYU422(_Interleaved422Base):
    @staticmethod
    def identifier():
        return "yvyu422"

    @staticmethod
    def bitdepth():
        return 8

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y0', "<u1"),
              ('v0', "<u1"),
              ('y1', "<u1"),
              ('u0', "<u1")],
             (self._height, self._width // 2))
        ])


pixel_formats.register(YUYV422)
pixel_formats.register(UYVY422)
pixel_formats.register(YVYU422)



from abc import ABC
import numpy as np
from .. import pixel_formats
from ..core import Format


class _YUV420IBase(Format, ABC):
    """Base for all interleaved yuv420 formats."""
    @staticmethod
    def chroma_subsampling():
        return 2, 2

    def unpack(self, data):
        even_lines = data['frame']['even_line']
        odd_lines = data['frame']['odd_line']

        frames = even_lines.shape[0]
        y = np.empty((frames, self._height, self._width), dtype=even_lines['y0'].dtype)

        y[:, 0::2, 0::2] = even_lines['y0']
        y[:, 0::2, 1::2] = even_lines['y1']
        y[:, 1::2, :] = odd_lines
        u = even_lines['u0']
        v = even_lines['v0']
        return y, u, v

    def pack(self, yuv):
        y, u, v = yuv
        data = np.empty(y.shape[0], dtype=self.dtype)
        data['frame']['even_line']['y0'][:] = y[:, 0::2, 0::2]
        data['frame']['even_line']['y1'][:] = y[:, 0::2, 1::2]
        data['frame']['odd_line'][:] = y[:, 1::2, :]
        data['frame']['even_line']['u0'][:] = u
        data['frame']['even_line']['v0'][:] = v
        return data


class YUV420I(_YUV420IBase):
    @staticmethod
    def identifier():
        return "yuv420i"

    @staticmethod
    def bitdepth():
        return 8

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [
                ('even_line', [('y0', "<u1"), ('u0', "<u1"), ('y1', "<u1"), ('v0', "<u1")], (self._width // 2)),
                ('odd_line', "<u1", self._width)
             ],
             (self._height // 2))
        ])


class YUV420I10LE(_YUV420IBase):
    @staticmethod
    def identifier():
        return "yuv420i10le"

    @staticmethod
    def bitdepth():
        return 10

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [
                ('even_line', [('y0', "<u2"), ('u0', "<u2"), ('y1', "<u2"), ('v0', "<u2")], (self._width // 2)),
                ('odd_line', "<u2", self._width)
             ],
             (self._height // 2))
        ])


class YUV420I10BE(_YUV420IBase):
    @staticmethod
    def identifier():
        return "yuv420i10be"

    @staticmethod
    def bitdepth():
        return 10

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [
                ('even_line', [('y0', ">u2"), ('u0', ">u2"), ('y1', ">u2"), ('v0', ">u2")], (self._width // 2)),
                ('odd_line', ">u2", self._width)
             ],
             (self._height // 2))
        ])


class YUV420I16LE(_YUV420IBase):
    @staticmethod
    def identifier():
        return "yuv420i16le"

    @staticmethod
    def bitdepth():
        return 16

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [
                ('even_line', [('y0', "<u2"), ('u0', "<u2"), ('y1', "<u2"), ('v0', "<u2")], (self._width // 2)),
                ('odd_line', "<u2", self._width)
             ],
             (self._height // 2))
        ])


class YUV420I16BE(_YUV420IBase):
    @staticmethod
    def identifier():
        return "yuv420i16be"

    @staticmethod
    def bitdepth():
        return 16

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [
                ('even_line', [('y0', ">u2"), ('u0', ">u2"), ('y1', ">u2"), ('v0', ">u2")], (self._width // 2)),
                ('odd_line', ">u2", self._width)
             ],
             (self._height // 2))
        ])


class YUV420I9LE(_YUV420IBase):
    @staticmethod
    def identifier():
        return "yuv420i9le"

    @staticmethod
    def bitdepth():
        return 9

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [
                ('even_line', [('y0', "<u2"), ('u0', "<u2"), ('y1', "<u2"), ('v0', "<u2")], (self._width // 2)),
                ('odd_line', "<u2", self._width)
             ],
             (self._height // 2))
        ])


class YUV420I9BE(_YUV420IBase):
    @staticmethod
    def identifier():
        return "yuv420i9be"

    @staticmethod
    def bitdepth():
        return 9

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [
                ('even_line', [('y0', ">u2"), ('u0', ">u2"), ('y1', ">u2"), ('v0', ">u2")], (self._width // 2)),
                ('odd_line', ">u2", self._width)
             ],
             (self._height // 2))
        ])


class YUV420I12LE(_YUV420IBase):
    @staticmethod
    def identifier():
        return "yuv420i12le"

    @staticmethod
    def bitdepth():
        return 12

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [
                ('even_line', [('y0', "<u2"), ('u0', "<u2"), ('y1', "<u2"), ('v0', "<u2")], (self._width // 2)),
                ('odd_line', "<u2", self._width)
             ],
             (self._height // 2))
        ])


class YUV420I12BE(_YUV420IBase):
    @staticmethod
    def identifier():
        return "yuv420i12be"

    @staticmethod
    def bitdepth():
        return 12

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [
                ('even_line', [('y0', ">u2"), ('u0', ">u2"), ('y1', ">u2"), ('v0', ">u2")], (self._width // 2)),
                ('odd_line', ">u2", self._width)
             ],
             (self._height // 2))
        ])


class YUV420I14LE(_YUV420IBase):
    @staticmethod
    def identifier():
        return "yuv420i14le"

    @staticmethod
    def bitdepth():
        return 14

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [
                ('even_line', [('y0', "<u2"), ('u0', "<u2"), ('y1', "<u2"), ('v0', "<u2")], (self._width // 2)),
                ('odd_line', "<u2", self._width)
             ],
             (self._height // 2))
        ])


class YUV420I14BE(_YUV420IBase):
    @staticmethod
    def identifier():
        return "yuv420i14be"

    @staticmethod
    def bitdepth():
        return 14

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [
                ('even_line', [('y0', ">u2"), ('u0', ">u2"), ('y1', ">u2"), ('v0', ">u2")], (self._width // 2)),
                ('odd_line', ">u2", self._width)
             ],
             (self._height // 2))
        ])


pixel_formats.register(YUV420I)
pixel_formats.register(YUV420I10LE)
pixel_formats.register(YUV420I10BE)
pixel_formats.register(YUV420I16LE)
pixel_formats.register(YUV420I16BE)
pixel_formats.register(YUV420I9LE)
pixel_formats.register(YUV420I9BE)
pixel_formats.register(YUV420I12LE)
pixel_formats.register(YUV420I12BE)
pixel_formats.register(YUV420I14LE)
pixel_formats.register(YUV420I14BE)



from .core import FormatManager
from .core import colorspaces

pixel_formats = FormatManager()

from .core.functions import imread, mimread, imwrite, mimwrite
from .core.functions import get_reader, get_writer
from .core.functions import frame, empty, zeros, ones, from_rgb
from .core.functions import to_rgb, from_rgb
from . import formats



from typing import Union
import io
from pathlib import Path
import psutil
import numpy as np
from . import YUVFrame
from . import Format


class Reader:

    def __init__(self, file: Union[Path, str, io.RawIOBase, io.BufferedIOBase], format: Format):
        if isinstance(file, io.RawIOBase) or isinstance(file, io.BufferedIOBase):
            self._owns = False
            self._file = file
        else:
            self._owns = True
            self._file = open(Path(file).expanduser().resolve(), 'rb')
            self._open = True
        self._format = format
        self._start_offset = self._file.tell()
        self._length = self._length_from_stream()
        self._iter_idx = 0

    def close(self):
        if not self._owns:
            raise RuntimeError("Cannot close unowned IO handle.")
        if not self._open:
            raise IOError("File is already closed.")
        self._file.close()
        self._open = False

    def __del__(self):
        if self._owns and self._open:
            self.close()

    def __len__(self):
        return self._length

    def __iter__(self):
        for i in range(self._length):
            yield self.read(i, count=1)[0]

    def _length_from_stream(self):
        self._file.seek(0, io.SEEK_END)
        num_bytes = self._file.tell() - self._start_offset
        self._file.seek(self._start_offset, io.SEEK_SET)
        return num_bytes // self._format.dtype.itemsize

    def _validate_memory(self, count):
        available = psutil.virtual_memory().available
        required = count * self._format.dtype.itemsize
        if required > available * 0.9:
            raise RuntimeError("The required memory ({}) to read '{}' frames "
                               "from file '{}' exceeds 90% of the available system "
                               "memory ({})".format(required,
                                                    count,
                                                    self._file.name,
                                                    available))

    def read(self, index, count=None):
        if count is None:
            count = self._length - index
        if index + count > self._length:
            raise ValueError("Cannot read number of frames '{}' at index '{}' "
                             "from file '{}' with length '{}'.".format(count,
                                                                       index,
                                                                       self._file.name,
                                                                       self._length))
        self._validate_memory(count)
        self._file.seek(self._start_offset + index * self._format.dtype.itemsize)
        data = np.empty(count, dtype=self._format.dtype)
        self._file.readinto(data.data)
        return self.unpack_data(data)

    def unpack_data(self, data):
        y_frames, u_frames, v_frames = self._format.unpack(data)

        yuv_frames = []
        if self._format.chroma_subsampling() != (0, 0):
            for y_frame, u_frame, v_frame in zip(y_frames, u_frames, v_frames):
                yuv_frames.append(YUVFrame(y_frame, u_frame, v_frame, self._format))
        else:
            for y_frame in y_frames:
                yuv_frames.append(YUVFrame(y_frame, None, None, self._format))
        return yuv_frames
